<?php

namespace App\Http\Controllers\API;

use App\Http\Controllers\Controller;
use App\Models\Member;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Hash;
use Illuminate\Support\Facades\Validator;
use Symfony\Component\HttpFoundation\Response;

class ApiAuthController extends Controller
{
    public function login(Request $request)
    {
        $validator = Validator::make($request->all(), [
            'phone_number' => 'required|max:255',
            'pin' => 'required|string',
        ],
            [
                'phone_number.required' => 'Your username is required',
                'phone_number.max' => 'Your username is too long and most probably incorrect',
                'pin.required' => 'A PIN is required to login',
            ]
        );

        if ($validator->fails()) {
            $response = [
                'status' => false,
                'message' => 'There were some problems with your input',
                'data' => $validator->errors()->all()
            ];
            return response($response, Response::HTTP_UNPROCESSABLE_ENTITY);
        }

        $user = Member::where('phone_number','=', $request->phone_number)->first();

        if ($user) {
            if (Hash::check($request->pin, $user->pin)) {
                $user->save();

                $token = $user->createToken('ZANUPF Password Grant Client');
                Auth::login($user);

                $response = [
                    'status' => true,
                    'message' => 'Welcome ' . $user->first_name,
                    'data' => [
                        'user' => $user,
                        'token' => $token,
                    ]
                ];
                return response($response, Response::HTTP_OK);
            } else {
                $response = ["status" => false, "message" => "Sorry, the details you have provided have not been recognised, please review and try again or contact our support team to get further assistance.",'data' => []];
                return response($response, Response::HTTP_UNPROCESSABLE_ENTITY);
            }
        } else {
            $response = ["status" => false, "message" => 'Sorry, you cannot login at this time. Your user was not found.','data' => []];
            return response($response, Response::HTTP_UNPROCESSABLE_ENTITY);
        }
    }

    public function register(Request $request)
    {
        // Validation
        $validated = $request->validate([
            'first_name' => 'required|string',
            'last_name' => 'required|string',
            'gender' => 'required|string|in:male,female,other',
            'date_of_birth' => 'required|date',
            'identification_number' => 'required|string',
            'identification_type_id' => 'required|integer',
            'phone_number' => 'required|string',
            'email' => 'nullable|email',
            'province_id' => 'nullable|integer|exists:provinces,id',
            'admin_district_id' => 'nullable|integer|exists:admin_districts,id',
            'party_district_id' => 'nullable|integer|exists:party_districts,id',
            'branch_id' => 'nullable|integer|exists:branches,id',
            'cell_id' => 'nullable|integer|exists:cells,id',
            'polling_station_id' => 'nullable|integer|exists:polling_stations,id',
            'wing_id' => 'required|integer|exists:wings,id',
            'card_fee' => 'required|numeric',
            'initial_subscription_fee' => 'required|numeric',
            'monthly_subscription_fee' => 'required|numeric',
            'pin' => 'required|confirmed|string',
        ]);

        $validated['pin'] = Hash::make($validated['pin']);

        // Create the member
        $member = Member::create($validated);

        $token = $member->createToken('ZANUPF Password Grant Client');
        Auth::login($member);

        $response = [
            'status' => true,
            'message' => 'Welcome ' . $member->first_name,
            'data' => [
                'member' => $member,
                'token' => $token,
            ]
        ];
        return response($response, Response::HTTP_CREATED);
    }

    public function getMyProfile(){
        $user = Member::where('id','=', auth()->user()->id)->first();

        if ($user) {
            $response = [
                'status' => true,
                'message' => 'Hey ' . $user->first_name,
                'data' => [
                    'user' => $user,
                ]
            ];
            return response($response, Response::HTTP_OK);
        } else {
            $response = ["status" => false, "message" => 'Sorry, your profile was not found.','data' => []];
            return response($response, Response::HTTP_NOT_FOUND);
        }
    }

    public function updateMyProfile(Request $request) {
        $validator = Validator::make(
            $request->all(), [
                'first_name' => 'required|string',
                'last_name' => 'required|string',
                'gender' => 'required|string|in:male,female,other',
                'date_of_birth' => 'required|date',
                'identification_number' => 'required|string',
                'identification_type_id' => 'required|integer',
                'phone_number' => 'required|string',
                'email' => 'nullable|email',
            ]
        );

        if ($validator->fails()) {
            return response([
                "status" => false,
                "message" => "There were some problems with your input",
                'data' => $validator->errors()->all()
            ], Response::HTTP_UNPROCESSABLE_ENTITY);
        }

        $user = Member::where('id', '=', auth()->user()->id)->first();

        $user->first_name = $request->input('first_name');
        $user->last_name = $request->input('last_name');
        $user->identification_number = $request->input('identification_number');
        $user->phone_number = $request->input('phone_number');
        $user->email = $request->input('email');
        $user->gender = $request->input('gender');

        $user->save();
        $user->refresh();

        $response = [
            'status' => true,
            'message' => 'Profile updated successfully.',
            'data' => [
                'member' => $user,
            ]
        ];
        return response($response, Response::HTTP_OK);
    }

    public function updateMyPassword(Request $request) {
        $validator = Validator::make(
            $request->all(), [
            'current_pin' => ['required'],
            'pin' => ['required', 'confirmed'],
        ],
            [
                'current_pin.required' => 'Your current PIN is required to process your request.',
                'pin.required' => 'A PIN is required to process your request.',
                'password.confirmed' => 'You need to confirm your PIN in order to process your request.'
            ]
        );

        if ($validator->fails()) {
            return response([
                "status" => false,
                "message" => "There were some problems with your input",
                'data' => $validator->errors()->all()
            ], Response::HTTP_UNPROCESSABLE_ENTITY);
        }

        $user = Member::where('id', '=', auth()->user()->id)->first();

        if (is_null($user)) {
            $response = [
                'status' => false,
                'message' => 'User not found or user is not active',
                'data' => []
            ];
            return response($response, Response::HTTP_UNPROCESSABLE_ENTITY);
        }

        if ($user) {
            if (Hash::check($request->current_password, $user->password)) {
                $user->pin = Hash::make($request->input('pin'));
                $user->save();
            } else {
                $response = ["status" => false, "message" => "Sorry, the details you have provided have not been recognised, please review and try again or contact our support team to get further assistance.",'data' => []];
                return response($response, Response::HTTP_UNPROCESSABLE_ENTITY);
            }
        } else {
            $response = ["status" => false, "message" => 'Sorry, the details you have provided have not been recognised, please review and try again or contact our support team to get further assistance.','data' => []];
            return response($response, Response::HTTP_UNPROCESSABLE_ENTITY);
        }

        $response = [
            'status' => true,
            'message' => 'PIN updated successfully.',
            'data' => []
        ];
        return response($response, Response::HTTP_OK);
    }

    public function requestAccountDeletion(Request $request) {

        $user = Member::where('id','=',auth()->user()->id)->first();

        $user->tokens->each(function ($token) {
            $token->delete();
        });

        $user->delete();

        $response = [
            'status' => true,
            'message' => 'Account deletion requested successful.',
            'data' => []
        ];
        return response($response, Response::HTTP_OK);
    }
}
